# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides matrix primitive representations."""

from typing import Union

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.misc.checks import check_ndarray_is_float_int
from ansys.geometry.core.typing import Real, RealSequence

DEFAULT_MATRIX33 = np.identity(3)
"""Default value of the 3x3 identity matrix for the ``Matrix33`` class."""

DEFAULT_MATRIX44 = np.identity(4)
"""Default value of the 4x4 identity matrix for the ``Matrix44`` class."""


class Matrix(np.ndarray):
    """Provides matrix representation.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence
        Matrix arguments as a :class:`np.ndarray <numpy.ndarray>` class.
    """

    def __new__(cls, input: np.ndarray | RealSequence):
        """Initialize ``Matrix`` class."""
        obj = np.asarray(input).view(cls)
        obj.setflags(write=False)

        if obj is None or obj.ndim != 2:
            raise ValueError("Matrix should only be a 2D array.")

        check_ndarray_is_float_int(obj)

        return obj

    def determinant(self) -> Real:
        """Get the determinant of the matrix."""
        if self.shape[0] != self.shape[1]:
            raise ValueError("The determinant is only defined for square matrices.")
        return np.linalg.det(self)

    def inverse(self) -> "Matrix":
        """Provide the inverse of the matrix."""
        det = self.determinant()
        if det <= 0:
            raise ValueError("The matrix cannot be inversed because its determinant is zero.")
        return np.linalg.inv(self)

    @check_input_types
    def __mul__(self, other: Union["Matrix", np.ndarray]) -> "Matrix":
        """Get the multiplication of the matrix."""
        if self.shape[1] != other.shape[0]:
            raise ValueError(
                f"The dimensions of the matrices {self.shape[1]} and {other.shape[0]} are not multipliable."  # noqa : E501
            )
        return np.matmul(self, other).view(Matrix)

    @check_input_types
    def __eq__(self, other: "Matrix") -> bool:
        """Equals operator for the ``Matrix`` class."""
        return np.array_equal(self, other)

    def __ne__(self, other: "Matrix") -> bool:
        """Not equals operator for the ``Matrix`` class."""
        return not self == other


class Matrix33(Matrix):
    """Provides 3x3 matrix representation.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence | Matrix, default: DEFAULT_MATRIX33
        Matrix arguments as a :class:`np.ndarray <numpy.ndarray>` class.
    """

    def __new__(cls, input: np.ndarray | RealSequence | Matrix = DEFAULT_MATRIX33):
        """Initialize the ``Matrix33`` class."""
        obj = Matrix(input).view(cls)
        if input is DEFAULT_MATRIX33:
            return obj

        if obj.shape != (3, 3):
            raise ValueError("Matrix33 should only be a 2D array of shape (3,3).")

        return obj


class Matrix44(Matrix):
    """Provides 4x4 matrix representation.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence | Matrix, default: DEFAULT_MATRIX44
        Matrix arguments as a :class:`np.ndarray <numpy.ndarray>` class.
    """

    def __new__(cls, input: np.ndarray | RealSequence | Matrix = DEFAULT_MATRIX44):
        """Initialize the ``Matrix44`` class."""
        obj = Matrix(input).view(cls)
        if input is DEFAULT_MATRIX44:
            return obj

        if obj.shape != (4, 4):
            raise ValueError("Matrix44 should only be a 2D array of shape (4,4).")

        return obj



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a trapezoid."""

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity
import pyvista as pv
from scipy.spatial.transform import Rotation as SpatialRotation

from ansys.geometry.core.logger import LOG
from ansys.geometry.core.math.constants import ZERO_POINT2D
from ansys.geometry.core.math.matrix import Matrix33
from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.segment import SketchSegment
from ansys.geometry.core.typing import Real


class Trapezoid(SketchFace):
    """Provides for modeling a 2D trapezoid.

    Parameters
    ----------
    base_width : ~pint.Quantity | Distance | Real
        Width of the lower base of the trapezoid.
    height : ~pint.Quantity | Distance | Real
        Height of the slot.
    base_angle : ~pint.Quantity | Distance | Real
        Angle for trapezoid generation. Represents the angle
        on the base of the trapezoid.
    base_asymmetric_angle : ~pint.Quantity | Angle | Real | None, default: None
        Asymmetrical angles on each side of the trapezoid.
        The default is ``None``, in which case the trapezoid is symmetrical. If
        provided, the trapezoid is asymmetrical and the right corner angle
        at the base of the trapezoid is set to the provided value.
    center: Point2D, default: ZERO_POINT2D
        Center point of the trapezoid.
    angle : ~pint.Quantity | Angle | Real, default: 0
        Placement angle for orientation alignment.

    Notes
    -----
    If an asymmetric base angle is defined, the base angle is
    applied to the left-most angle, and the asymmetric base angle
    is applied to the right-most angle.
    """

    @check_input_types
    def __init__(
        self,
        base_width: Quantity | Distance | Real,
        height: Quantity | Distance | Real,
        base_angle: Quantity | Angle | Real,
        base_asymmetric_angle: Quantity | Angle | Real | None = None,
        center: Point2D = ZERO_POINT2D,
        angle: Quantity | Angle | Real = 0,
    ):
        """Initialize the trapezoid."""
        super().__init__()

        # TODO: Remove this warning in the next major release (v0.8.0)
        # https://github.com/ansys/pyansys-geometry/issues/1359
        LOG.warning(
            "The signature of the Trapezoid class has changed starting on "
            "version 0.7.X. Please refer to the documentation for more information."
        )

        self._center = center
        self._base_width = (
            base_width if isinstance(base_width, Distance) else Distance(base_width, center.unit)
        )
        if self._base_width.value <= 0:
            raise ValueError("Base width must be a real positive value.")
        width_magnitude = self._base_width.value.m_as(center.unit)

        self._height = height if isinstance(height, Distance) else Distance(height, center.unit)
        if self._height.value <= 0:
            raise ValueError("Height must be a real positive value.")
        height_magnitude = self._height.value.m_as(center.unit)

        if isinstance(angle, (int, float)):
            angle = Angle(angle, DEFAULT_UNITS.ANGLE)
        angle = angle if isinstance(angle, Angle) else Angle(angle, angle.units)

        if isinstance(base_angle, (int, float)):
            base_angle = Angle(base_angle, DEFAULT_UNITS.ANGLE)
        base_angle = (
            base_angle if isinstance(base_angle, Angle) else Angle(base_angle, base_angle.units)
        )

        if base_asymmetric_angle is None:
            base_asymmetric_angle = base_angle
        else:
            if isinstance(base_asymmetric_angle, (int, float)):
                base_asymmetric_angle = Angle(base_asymmetric_angle, DEFAULT_UNITS.ANGLE)
            base_asymmetric_angle = (
                base_asymmetric_angle
                if isinstance(base_asymmetric_angle, Angle)
                else Angle(base_asymmetric_angle, base_asymmetric_angle.units)
            )

        # SANITY CHECK: Ensure that the angles are valid (i.e. between 0 and 180 degrees)
        for trapz_angle in [base_angle, base_asymmetric_angle]:
            if (
                trapz_angle.value.m_as(UNITS.radian) < 0
                or trapz_angle.value.m_as(UNITS.radian) > np.pi
            ):
                raise ValueError("The trapezoid angles must be between 0 and 180 degrees.")

        # Check that the sum of both angles is larger than 90 degrees
        base_offset_right = height_magnitude / np.tan(
            base_asymmetric_angle.value.m_as(UNITS.radian)
        )
        base_offset_left = height_magnitude / np.tan(base_angle.value.m_as(UNITS.radian))

        # SANITY CHECK: Ensure that the trapezoid is not degenerate
        if base_offset_right + base_offset_left >= width_magnitude:
            raise ValueError(
                "The trapezoid is degenerate. "
                "The provided angles, width and height do not form a valid trapezoid."
            )

        rotation = Matrix33(
            SpatialRotation.from_euler(
                "xyz", [0, 0, angle.value.m_as(UNITS.radian)], degrees=False
            ).as_matrix()
        )

        half_h = height_magnitude / 2
        half_w = width_magnitude / 2
        rotated_point_1 = rotation @ [center.x.m - half_w, center.y.m - half_h, 0]
        rotated_point_2 = rotation @ [center.x.m + half_w, center.y.m - half_h, 0]
        rotated_point_3 = rotation @ [
            center.x.m + half_w - base_offset_right,
            center.y.m + half_h,
            0,
        ]
        rotated_point_4 = rotation @ [
            center.x.m - half_w + base_offset_left,
            center.y.m + half_h,
            0,
        ]

        self._point1 = Point2D([rotated_point_1[0], rotated_point_1[1]], center.unit)
        self._point2 = Point2D([rotated_point_2[0], rotated_point_2[1]], center.unit)
        self._point3 = Point2D([rotated_point_3[0], rotated_point_3[1]], center.unit)
        self._point4 = Point2D([rotated_point_4[0], rotated_point_4[1]], center.unit)

        # TODO: add plane to SketchSegment when available
        # https://github.com/ansys/pyansys-geometry/issues/1319
        self._segment1 = SketchSegment(self._point1, self._point2)
        self._segment2 = SketchSegment(self._point2, self._point3)
        self._segment3 = SketchSegment(self._point3, self._point4)
        self._segment4 = SketchSegment(self._point4, self._point1)

        self._edges.append(self._segment1)
        self._edges.append(self._segment2)
        self._edges.append(self._segment3)
        self._edges.append(self._segment4)

    @property
    def center(self) -> Point2D:
        """Center of the trapezoid."""
        return self._center

    @property
    def base_width(self) -> Quantity:
        """Width of the trapezoid."""
        return self._base_width.value

    @property
    def height(self) -> Quantity:
        """Height of the trapezoid."""
        return self._height.value

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        import numpy as np

        return pv.Quadrilateral(
            np.array(
                [
                    [
                        self._point1.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._point1.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self._point2.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._point2.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self._point3.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._point3.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self._point4.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._point4.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                ],
                dtype=np.float64,
            )
        )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a beam."""

from typing import TYPE_CHECKING

from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.checks import check_type
from ansys.geometry.core.misc.measurements import Distance

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.component import Component


class BeamProfile:
    """Represents a single beam profile organized within the design assembly.

    This profile synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the beam profile.
    name : str
        User-defined label for the beam profile.

    Notes
    -----
    ``BeamProfile`` objects are expected to be created from the ``Design`` object.
    This means that you are not expected to instantiate your own ``BeamProfile``
    object. You should call the specific ``Design`` API for the ``BeamProfile`` desired.
    """

    def __init__(self, id: str, name: str):
        """Initialize ``BeamProfile`` class."""
        self._id = id
        self._name = name

    @property
    def id(self) -> str:
        """ID of the beam profile."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the beam profile."""
        return self._name


class BeamCircularProfile(BeamProfile):
    """Represents a single circular beam profile.

    This profile synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the beam profile.
    name : str
        User-defined label for the beam profile.
    radius : Distance
        Radius of the circle.
    center: Point3D
        3D point representing the center of the circle.
    direction_x: UnitVector3D
        X-axis direction.
    direction_y: UnitVector3D
        Y-axis direction.

    Notes
    -----
    ``BeamProfile`` objects are expected to be created from the ``Design`` object.
    This means that you are not expected to instantiate your own ``BeamProfile``
    object. You should call the specific ``Design`` API for the ``BeamProfile`` desired.
    """

    def __init__(
        self,
        id: str,
        name: str,
        radius: Distance,
        center: Point3D,
        direction_x: UnitVector3D,
        direction_y: UnitVector3D,
    ):
        """Initialize ``BeamCircularProfile`` class."""
        super().__init__(id, name)

        # Store specific BeamCircularProfile variables
        self._radius = radius
        self._center = center
        self._dir_x = direction_x
        self._dir_y = direction_y

    @property
    def radius(self) -> Distance:
        """Radius of the circular beam profile."""
        return self._radius

    @property
    def center(self) -> Point3D:
        """Center of the circular beam profile."""
        return self._center

    @property
    def direction_x(self) -> UnitVector3D:
        """X-axis direction of the circular beam profile."""
        return self._dir_x

    @property
    def direction_y(self) -> UnitVector3D:
        """Y-axis direction of the circular beam profile."""
        return self._dir_y

    def __repr__(self) -> str:
        """Represent the ``BeamCircularProfile`` as a string."""
        lines = [f"ansys.geometry.core.designer.BeamCircularProfile {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Radius               : {str(self.radius.value)}")
        lines.append(
            f"  Center               : [{','.join([str(x) for x in self.center])}] in meters"
        )
        lines.append(f"  Direction x          : [{','.join([str(x) for x in self.direction_x])}]")
        lines.append(f"  Direction y          : [{','.join([str(x) for x in self.direction_y])}]")
        return "\n".join(lines)


class Beam:
    """Represents a simplified solid body with an assigned 2D cross-section.

    This body synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    name : str
        User-defined label for the body.
    start : Point3D
        Start of the beam line segment.
    end : Point3D
        End of the beam line segment.
    profile : BeamProfile
        Beam profile to use to create the beam.
    parent_component : Component
        Parent component to nest the new beam under within the design assembly.
    """

    def __init__(
        self,
        id: str,
        start: Point3D,
        end: Point3D,
        profile: BeamProfile,
        parent_component: "Component",
    ):
        """Initialize ``Beam`` class."""
        from ansys.geometry.core.designer.component import Component

        check_type(id, str)
        check_type(start, Point3D)
        check_type(end, Point3D)
        check_type(profile, BeamProfile)
        check_type(parent_component, Component)

        self._id = id
        self._start = start
        self._end = end
        self._profile = profile
        self._parent_component = parent_component
        self._is_alive = True

    @property
    def id(self) -> str:
        """Service-defined ID of the beam."""
        return self._id

    @property
    def start(self) -> Point3D:
        """Start of the beam line segment."""
        return self._start

    @property
    def end(self) -> Point3D:
        """End of the beam line segment."""
        return self._end

    @property
    def profile(self) -> BeamProfile:
        """Beam profile of the beam line segment."""
        return self._profile

    @property
    def parent_component(self) -> "Component":
        """Component node that the beam is under."""
        return self._parent_component

    @property
    def is_alive(self) -> bool:
        """Flag indicating whether the beam is still alive on the server."""
        return self._is_alive

    def __repr__(self) -> str:
        """Represent the beam as a string."""
        lines = [f"ansys.geometry.core.designer.Beam {hex(id(self))}"]
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(
            f"  Start                : [{','.join([str(x) for x in self.start])}] in meters"
        )
        lines.append(f"  End                  : [{','.join([str(x) for x in self.end])}] in meters")
        lines.append(f"  Parent component     : {self.parent_component.name}")
        lines.extend(["\n", "  Beam Profile info", "  -----------------", str(self.profile)])
        return "\n".join(lines)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing a frame."""

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Y, ZERO_POINT3D
from ansys.geometry.core.math.matrix import Matrix33, Matrix44
from ansys.geometry.core.math.point import Point2D, Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.typing import RealSequence


class Frame:
    """Representation of a frame.

    Parameters
    ----------
    origin : ~numpy.ndarray | RealSequence | Point3D, default: ZERO_POINT3D
        Centered origin of the`frame. The default is ``ZERO_POINT3D``, which is the
        Cartesian origin.
    direction_x : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D, default: UNITVECTOR3D_X
        X-axis direction.
    direction_y : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D, default: UNITVECTOR3D_Y
        Y-axis direction.
    """  # noqa : E501

    @check_input_types
    def __init__(
        self,
        origin: np.ndarray | RealSequence | Point3D = ZERO_POINT3D,
        direction_x: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_X,
        direction_y: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_Y,
    ):
        """Initialize the ``Frame`` class."""
        self._origin = Point3D(origin) if not isinstance(origin, Point3D) else origin
        self._direction_x = (
            UnitVector3D(direction_x) if not isinstance(direction_x, UnitVector3D) else direction_x
        )
        self._direction_y = (
            UnitVector3D(direction_y) if not isinstance(direction_y, UnitVector3D) else direction_y
        )

        # origin is fixed once the frame is built
        self._origin.setflags(write=False)

        if not self._direction_x.is_perpendicular_to(self._direction_y):
            raise ValueError("Direction x and direction y must be perpendicular.")

        self._direction_z = UnitVector3D(self._direction_x % self._direction_y)

        self._rotation_matrix = Matrix33(
            np.array(
                [
                    self.direction_x.tolist(),
                    self.direction_y.tolist(),
                    self.direction_z.tolist(),
                ]
            )
        )

        self._transformation_matrix = Matrix44(
            [
                [
                    self.direction_x.x,
                    self.direction_y.x,
                    self.direction_z.x,
                    self.origin.x.to_base_units().m,
                ],
                [
                    self.direction_x.y,
                    self.direction_y.y,
                    self.direction_z.y,
                    self.origin.y.to_base_units().m,
                ],
                [
                    self.direction_x.z,
                    self.direction_y.z,
                    self.direction_z.z,
                    self.origin.z.to_base_units().m,
                ],
                [0, 0, 0, 1],
            ]
        )

    @property
    def origin(self) -> Point3D:
        """Origin of the frame."""
        return self._origin

    @property
    def direction_x(self) -> UnitVector3D:
        """X-axis direction of the frame."""
        return self._direction_x

    @property
    def direction_y(self) -> UnitVector3D:
        """Y-axis direction of the frame."""
        return self._direction_y

    @property
    def direction_z(self) -> UnitVector3D:
        """Z-axis direction of the frame."""
        return self._direction_z

    @property
    def global_to_local_rotation(self) -> Matrix33:
        """Global to local space transformation matrix.

        Returns
        -------
        Matrix33
            3x3 matrix representing the transformation from global to local
            coordinate space, excluding origin translation.
        """
        return self._rotation_matrix

    @property
    def local_to_global_rotation(self) -> Matrix33:
        """Local to global space transformation matrix.

        Returns
        -------
        Matrix33
            3x3 matrix representing the transformation from local to global
            coordinate space.
        """
        return self._rotation_matrix.T

    @property
    def transformation_matrix(self) -> Matrix44:
        """Full 4x4 transformation matrix.

        Returns
        -------
        Matrix44
            4x4 matrix representing the transformation from global to local
            coordinate space.
        """
        return self._transformation_matrix

    @check_input_types
    def transform_point2d_local_to_global(self, point: Point2D) -> Point3D:
        """Transform a 2D point to a global 3D point.

        This method transforms a local, plane-contained ``Point2D`` object in the global
        coordinate system, thus representing it as a ``Point3D`` object.

        Parameters
        ----------
        point : Point2D
            ``Point2D`` local object to express in global coordinates.

        Returns
        -------
        Point3D
            Global coordinates for the 3D point.
        """
        return self.origin + Vector3D(self.local_to_global_rotation @ [point[0], point[1], 0])

    @check_input_types
    def __eq__(self, other: "Frame") -> bool:
        """Equals operator for the ``Frame`` class."""
        return (
            self.origin == other.origin
            and self.direction_x == other.direction_x
            and self.direction_y == other.direction_y
            and self.direction_z == other.direction_z
        )

    def __ne__(self, other: "Frame") -> bool:
        """Not equals operator for the ``Frame`` class."""
        return not self == other



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for managing a face."""

from enum import Enum, unique
from typing import TYPE_CHECKING

from pint import Quantity

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.api.geometry.v0.edges_pb2_grpc import EdgesStub
from ansys.api.geometry.v0.faces_pb2 import (
    CreateIsoParamCurvesRequest,
    EvaluateRequest,
    GetNormalRequest,
)
from ansys.api.geometry.v0.faces_pb2_grpc import FacesStub
from ansys.api.geometry.v0.models_pb2 import Edge as GRPCEdge
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.connection.conversions import grpc_curve_to_curve, grpc_surface_to_surface
from ansys.geometry.core.designer.edge import Edge
from ansys.geometry.core.errors import GeometryRuntimeError, protect_grpc
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.checks import (
    deprecated_method,
    ensure_design_is_active,
    min_backend_version,
)
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS
from ansys.geometry.core.shapes.box_uv import BoxUV
from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
from ansys.geometry.core.shapes.parameterization import Interval
from ansys.geometry.core.shapes.surfaces.trimmed_surface import (
    ReversedTrimmedSurface,
    TrimmedSurface,
)

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body


@unique
class SurfaceType(Enum):
    """Provides values for the surface types supported."""

    SURFACETYPE_UNKNOWN = 0
    SURFACETYPE_PLANE = 1
    SURFACETYPE_CYLINDER = 2
    SURFACETYPE_CONE = 3
    SURFACETYPE_TORUS = 4
    SURFACETYPE_SPHERE = 5
    SURFACETYPE_NURBS = 6
    SURFACETYPE_PROCEDURAL = 7


@unique
class FaceLoopType(Enum):
    """Provides values for the face loop types supported."""

    INNER_LOOP = "INNER"
    OUTER_LOOP = "OUTER"


class FaceLoop:
    """Provides an internal class holding the face loops defined.

    Notes
    -----
    This class is to be used only when parsing server side results. It is not
    intended to be instantiated by a user.

    Parameters
    ----------
    type : FaceLoopType
        Type of loop.
    length : Quantity
        Length of the loop.
    min_bbox : Point3D
        Minimum point of the bounding box containing the loop.
    max_bbox : Point3D
        Maximum point of the bounding box containing the loop.
    edges : list[Edge]
        Edges contained in the loop.
    """

    def __init__(
        self,
        type: FaceLoopType,
        length: Quantity,
        min_bbox: Point3D,
        max_bbox: Point3D,
        edges: list[Edge],
    ):
        """Initialize ``FaceLoop`` class."""
        self._type = type
        self._length = length
        self._min_bbox = min_bbox
        self._max_bbox = max_bbox
        self._edges = edges

    @property
    def type(self) -> FaceLoopType:
        """Type of the loop."""
        return self._type

    @property
    def length(self) -> Quantity:
        """Length of the loop."""
        return self._length

    @property
    def min_bbox(self) -> Point3D:
        """Minimum point of the bounding box containing the loop."""
        return self._min_bbox

    @property
    def max_bbox(self) -> Point3D:
        """Maximum point of the bounding box containing the loop."""
        return self._max_bbox

    @property
    def edges(self) -> list[Edge]:
        """Edges contained in the loop."""
        return self._edges


class Face:
    """Represents a single face of a body within the design assembly.

    This class synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    surface_type : SurfaceType
        Type of surface that the face forms.
    body : Body
        Parent body that the face constructs.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    """

    def __init__(
        self,
        id: str,
        surface_type: SurfaceType,
        body: "Body",
        grpc_client: GrpcClient,
        is_reversed: bool = False,
    ):
        """Initialize the ``Face`` class."""
        self._id = id
        self._surface_type = surface_type
        self._body = body
        self._grpc_client = grpc_client
        self._faces_stub = FacesStub(grpc_client.channel)
        self._edges_stub = EdgesStub(grpc_client.channel)
        self._is_reversed = is_reversed
        self._shape = None

        self._grpc_client.log.debug("Requesting surface properties from server.")

    @property
    def id(self) -> str:
        """Face ID."""
        return self._id

    @property
    def _grpc_id(self) -> EntityIdentifier:
        """Entity ID of this face on the server side."""
        return EntityIdentifier(id=self._id)

    @property
    def is_reversed(self) -> bool:
        """Flag indicating if the face is reversed."""
        return self._is_reversed

    @property
    def body(self) -> "Body":
        """Body that the face belongs to."""
        return self._body

    @property
    @protect_grpc
    @ensure_design_is_active
    @min_backend_version(24, 2, 0)
    def shape(self) -> TrimmedSurface:
        """Underlying trimmed surface of the face.

        If the face is reversed, its shape is a ``ReversedTrimmedSurface`` type, which handles the
        direction of the normal vector to ensure it is always facing outward.
        """
        if self._shape is None:
            self._grpc_client.log.debug("Requesting face properties from server.")

            surface_response = self._faces_stub.GetSurface(self._grpc_id)
            geometry = grpc_surface_to_surface(surface_response, self._surface_type)
            box = self._faces_stub.GetBoxUV(self._grpc_id)
            box_uv = BoxUV(Interval(box.start_u, box.end_u), Interval(box.start_v, box.end_v))

            self._shape = (
                ReversedTrimmedSurface(geometry, box_uv)
                if self.is_reversed
                else TrimmedSurface(geometry, box_uv)
            )
        return self._shape

    @property
    def surface_type(self) -> SurfaceType:
        """Surface type of the face."""
        return self._surface_type

    @property
    @protect_grpc
    @ensure_design_is_active
    def area(self) -> Quantity:
        """Calculated area of the face."""
        self._grpc_client.log.debug("Requesting face area from server.")
        area_response = self._faces_stub.GetArea(self._grpc_id)
        return Quantity(area_response.area, DEFAULT_UNITS.SERVER_AREA)

    @property
    @protect_grpc
    @ensure_design_is_active
    def edges(self) -> list[Edge]:
        """List of all edges of the face."""
        self._grpc_client.log.debug("Requesting face edges from server.")
        edges_response = self._faces_stub.GetEdges(self._grpc_id)
        return self.__grpc_edges_to_edges(edges_response.edges)

    @property
    @protect_grpc
    @ensure_design_is_active
    def loops(self) -> list[FaceLoop]:
        """List of all loops of the face."""
        self._grpc_client.log.debug("Requesting face loops from server.")
        grpc_loops = self._faces_stub.GetLoops(EntityIdentifier(id=self.id)).loops
        loops = []
        for grpc_loop in grpc_loops:
            type = FaceLoopType(grpc_loop.type)
            length = Quantity(grpc_loop.length, DEFAULT_UNITS.SERVER_LENGTH)
            min = Point3D(
                [
                    grpc_loop.bounding_box.min.x,
                    grpc_loop.bounding_box.min.y,
                    grpc_loop.bounding_box.min.z,
                ],
                DEFAULT_UNITS.SERVER_LENGTH,
            )
            max = Point3D(
                [
                    grpc_loop.bounding_box.max.x,
                    grpc_loop.bounding_box.max.y,
                    grpc_loop.bounding_box.max.z,
                ],
                DEFAULT_UNITS.SERVER_LENGTH,
            )
            grpc_edges = [
                self._edges_stub.Get(EntityIdentifier(id=edge_id)) for edge_id in grpc_loop.edges
            ]
            edges = self.__grpc_edges_to_edges(grpc_edges)
            loops.append(
                FaceLoop(type=type, length=length, min_bbox=min, max_bbox=max, edges=edges)
            )

        return loops

    @protect_grpc
    @ensure_design_is_active
    def normal(self, u: float = 0.5, v: float = 0.5) -> UnitVector3D:
        """Get the normal direction to the face at certain UV coordinates.

        Notes
        -----
        To properly use this method, you must handle UV coordinates. Thus, you must
        know how these relate to the underlying Geometry service. It is an advanced
        method for Geometry experts only.

        Parameters
        ----------
        u : float, default: 0.5
            First coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.
        v : float, default: 0.5
            Second coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.

        Returns
        -------
        UnitVector3D
            :class:`UnitVector3D` object evaluated at the given U and V coordinates.
            This :class:`UnitVector3D` object is perpendicular to the surface at the
            given UV coordinates.
        """
        try:
            return self.shape.normal(u, v)
        except GeometryRuntimeError:  # pragma: no cover
            # Only for versions earlier than 24.2.0 (before the introduction of the shape property)
            self._grpc_client.log.debug(f"Requesting face normal from server with (u,v)=({u},{v}).")
            response = self._faces_stub.GetNormal(GetNormalRequest(id=self.id, u=u, v=v)).direction
            return UnitVector3D([response.x, response.y, response.z])

    @deprecated_method(alternative="normal")
    def face_normal(self, u: float = 0.5, v: float = 0.5) -> UnitVector3D:  # [deprecated-method]
        """Get the normal direction to the face at certain UV coordinates.

        Notes
        -----
        This method is deprecated. Use the ``normal`` method instead.

        Parameters
        ----------
        u : float, default: 0.5
            First coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.
        v : float, default: 0.5
            Second coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.

        Returns
        -------
        UnitVector3D
            :class:`UnitVector3D` object evaluated at the given U and V coordinates.
            This :class:`UnitVector3D` object is perpendicular to the surface at the
            given UV coordinates.
        """
        return self.normal(u, v)

    @protect_grpc
    @ensure_design_is_active
    def point(self, u: float = 0.5, v: float = 0.5) -> Point3D:
        """Get a point of the face evaluated at certain UV coordinates.

        Notes
        -----
        To properly use this method, you must handle UV coordinates. Thus, you must
        know how these relate to the underlying Geometry service. It is an advanced
        method for Geometry experts only.

        Parameters
        ----------
        u : float, default: 0.5
            First coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.
        v : float, default: 0.5
            Second coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.

        Returns
        -------
        Point3D
            :class:`Point3D` object evaluated at the given UV coordinates.
        """
        try:
            return self.shape.evaluate_proportion(u, v).position
        except GeometryRuntimeError:  # pragma: no cover
            # Only for versions earlier than 24.2.0 (before the introduction of the shape property)
            self._grpc_client.log.debug(f"Requesting face point from server with (u,v)=({u},{v}).")
            response = self._faces_stub.Evaluate(EvaluateRequest(id=self.id, u=u, v=v)).point
            return Point3D([response.x, response.y, response.z], DEFAULT_UNITS.SERVER_LENGTH)

    @deprecated_method(alternative="point")
    def face_point(self, u: float = 0.5, v: float = 0.5) -> Point3D:
        """Get a point of the face evaluated at certain UV coordinates.

        Notes
        -----
        This method is deprecated. Use the ``point`` method instead.

        Parameters
        ----------
        u : float, default: 0.5
            First coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.
        v : float, default: 0.5
            Second coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.

        Returns
        -------
        Point3D
            :class:`Point3D` object evaluated at the given UV coordinates.
        """
        return self.point(u, v)

    def __grpc_edges_to_edges(self, edges_grpc: list[GRPCEdge]) -> list[Edge]:
        """Transform a list of gRPC edge messages into actual ``Edge`` objects.

        Parameters
        ----------
        edges_grpc : list[GRPCEdge]
            list of gRPC messages of type ``Edge``.

        Returns
        -------
        list[Edge]
            ``Edge`` objects to obtain from gRPC messages.
        """
        from ansys.geometry.core.designer.edge import CurveType, Edge

        edges = []
        for edge_grpc in edges_grpc:
            edges.append(
                Edge(
                    edge_grpc.id,
                    CurveType(edge_grpc.curve_type),
                    self._body,
                    self._grpc_client,
                    edge_grpc.is_reversed,
                )
            )
        return edges

    @protect_grpc
    @ensure_design_is_active
    def create_isoparametric_curves(
        self, use_u_param: bool, parameter: float
    ) -> list[TrimmedCurve]:
        """Create isoparametic curves at the given proportional parameter.

        Typically, only one curve is created, but if the face has a hole, it is possible that
        more than one curve is created.

        Parameters
        ----------
        use_u_param : bool
            Whether the parameter is the ``u`` coordinate or ``v`` coordinate. If ``True``,
            it is the ``u`` coordinate. If ``False``, it is the ``v`` coordinate.
        parameter : float
            Proportional [0-1] parameter to create the one or more curves at.

        Returns
        -------
        list[TrimmedCurve]
            list of curves that were created.
        """
        curves = self._faces_stub.CreateIsoParamCurves(
            CreateIsoParamCurvesRequest(id=self.id, u_dir_curve=use_u_param, proportion=parameter)
        ).curves

        trimmed_curves = []
        for c in curves:
            geometry = grpc_curve_to_curve(c.curve)
            start = Point3D([c.start.x, c.start.y, c.start.z])
            end = Point3D([c.end.x, c.end.y, c.end.z])
            interval = Interval(c.interval_start, c.interval_end)
            length = Quantity(c.length, DEFAULT_UNITS.SERVER_LENGTH)

            trimmed_curves.append(
                TrimmedCurve(geometry, start, end, interval, length, self._grpc_client)
            )

        return trimmed_curves



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ``Curve`` class."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.shapes.curves.curve_evaluation import CurveEvaluation
from ansys.geometry.core.shapes.parameterization import Interval, Parameterization
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve


class Curve(ABC):
    """Provides the abstract base class representing a 3D curve."""

    @abstractmethod
    def parameterization(self) -> Parameterization:
        """Parameterize the curve."""
        return

    @abstractmethod
    def contains_param(self, param: Real) -> bool:
        """Check a parameter is within the parametric range of the curve."""
        return

    @abstractmethod
    def contains_point(self, point: Point3D) -> bool:
        """Check a point is contained by the curve.

        The point can either lie within the curve or on its boundary.
        """
        return

    @abstractmethod
    def transformed_copy(self, matrix: Matrix44) -> "Curve":
        """Create a transformed copy of the curve."""
        return

    @abstractmethod
    def __eq__(self, other: "Curve") -> bool:
        """Determine if two curves are equal."""
        return

    @abstractmethod
    def evaluate(self, parameter: Real) -> CurveEvaluation:
        """Evaluate the curve at the given parameter."""
        return

    @abstractmethod
    def project_point(self, point: Point3D) -> CurveEvaluation:
        """Project a point to the curve.

        This method returns the evaluation at the closest point.
        """
        return

    def trim(self, interval: Interval) -> "TrimmedCurve":
        """Trim this curve by bounding it with an interval.

        Returns
        -------
        TrimmedCurve
            The resulting bounded curve.
        """
        from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve

        return TrimmedCurve(
            self,
            self.evaluate(interval.start).position,
            self.evaluate(interval.end).position,
            interval,
            None,
        )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides primitive representation of a 2D plane in 3D space."""

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Y, ZERO_POINT3D
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.typing import RealSequence


class Plane(Frame):
    """Provides primitive representation of a 2D plane in 3D space.

    Parameters
    ----------
    origin : ~numpy.ndarray | RealSequence | Point3D, default: ZERO_POINT3D
        Centered origin of the frame. The default is ``ZERO_POINT3D``, which is the
        Cartesian origin.
    direction_x : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D, default: UNITVECTOR3D_X
        X-axis direction.
    direction_y : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D, default: UNITVECTOR3D_Y
        Y-axis direction.
    """  # noqa : E501

    def __init__(
        self,
        origin: np.ndarray | RealSequence | Point3D = ZERO_POINT3D,
        direction_x: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_X,
        direction_y: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_Y,
    ):
        """Initialize ``Plane`` class."""
        super().__init__(origin, direction_x, direction_y)

    @check_input_types
    def is_point_contained(self, point: Point3D) -> bool:
        """Check if a 3D point is contained in the plane.

        Parameters
        ----------
        point : Point3D
            :class:`Point3D <ansys.geometry.core.math.point.Point3D>` class to check.

        Returns
        -------
        bool
            ``True`` if the 3D point is contained in the plane, ``False`` otherwise.
        """
        # Compute the plane equation A*(x-x0) + B*(y-y0) + C*(z-z0)
        plane_eq = (
            self.direction_z.x * (point.x - self.origin.x).m
            + self.direction_z.y * (point.y - self.origin.y).m
            + self.direction_z.z * (point.z - self.origin.z).m
        )

        # If plane equation is equal to 0, your point is contained
        return True if np.isclose(plane_eq, 0.0) else False

    @property
    def normal(self) -> UnitVector3D:
        """Calculate the normal vector of the plane.

        Returns
        -------
        UnitVector3D
            Normal vector of the plane.
        """
        return self.direction_z

    @check_input_types
    def __eq__(self, other: "Plane") -> bool:
        """Equals operator for the ``Plane`` class."""
        return (
            self.origin == other.origin
            and self.direction_x == other.direction_x
            and self.direction_y == other.direction_y
            and self.direction_z == other.direction_z
        )

    def __ne__(self, other: "Plane") -> bool:
        """Not equals operator for the ``Plane`` class."""
        return not self == other



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a plane."""

from functools import cached_property

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Z
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
    ParamUV,
)
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.surface_evaluation import SurfaceEvaluation
from ansys.geometry.core.typing import Real, RealSequence


class PlaneSurface(Surface):
    """Provides 3D plane surface representation.

    Parameters
    ----------
    origin : ~numpy.ndarray | RealSequence | Point3D
        Centered origin of the plane.
    reference : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
        X-axis direction.
    axis : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
        X-axis direction.
    """

    def __init__(
        self,
        origin: np.ndarray | RealSequence | Point3D,
        reference: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_X,
        axis: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_Z,
    ):
        """Initialize an instance of a plane surface."""
        self._origin = Point3D(origin) if not isinstance(origin, Point3D) else origin

        self._reference = (
            UnitVector3D(reference) if not isinstance(reference, UnitVector3D) else reference
        )
        self._axis = UnitVector3D(axis) if not isinstance(axis, UnitVector3D) else axis
        if not self._reference.is_perpendicular_to(self._axis):
            raise ValueError("Plane reference (dir_x) and axis (dir_z) must be perpendicular.")

    @property
    def origin(self) -> Point3D:
        """Origin of the plane."""
        return self._origin

    @property
    def dir_x(self) -> UnitVector3D:
        """X-direction of the plane."""
        return self._reference

    @property
    def dir_y(self) -> UnitVector3D:
        """Y-direction of the plane."""
        return self.dir_z.cross(self.dir_x)

    @property
    def dir_z(self) -> UnitVector3D:
        """Z-direction of the plane."""
        return self._axis

    @check_input_types
    def __eq__(self, other: "PlaneSurface") -> bool:
        """Check whether two planes are equal."""
        return (
            self._origin == other._origin
            and self._reference == other._reference
            and self._axis == other._axis
        )

    def contains_param(self, param_uv: ParamUV) -> bool:
        """Check a ``ParamUV`` is within the parametric range of the surface."""
        raise NotImplementedError("contains_param() is not implemented.")

    def contains_point(self, point: Point3D) -> bool:
        """Check whether a 3D point is in the domain of the plane."""
        raise NotImplementedError("contains_point() is not implemented.")

    def parameterization(self) -> tuple[Parameterization, Parameterization]:
        """Parametrize the plane."""
        u = Parameterization(ParamForm.OPEN, ParamType.LINEAR, Interval(-np.inf, np.inf))
        v = Parameterization(ParamForm.OPEN, ParamType.LINEAR, Interval(-np.inf, np.inf))

        return (u, v)

    def project_point(self, point: Point3D) -> SurfaceEvaluation:
        """Evaluate the plane at a given 3D point."""
        origin_to_point = point - self._origin
        u = origin_to_point.dot(self.dir_x)
        v = origin_to_point.dot(self.dir_y)
        return PlaneEvaluation(self, ParamUV(u, v))

    def transformed_copy(self, matrix: Matrix44) -> Surface:
        """Get a transformed version of the plane given the transform."""
        new_point = self.origin.transform(matrix)
        new_reference = self._reference.transform(matrix)
        new_axis = self._axis.transform(matrix)
        return PlaneSurface(
            new_point,
            UnitVector3D(new_reference[0:3]),
            UnitVector3D(new_axis[0:3]),
        )

    def evaluate(self, parameter: ParamUV) -> "PlaneEvaluation":
        """Evaluate the plane at a given u and v parameter."""
        return PlaneEvaluation(self, parameter)


class PlaneEvaluation(SurfaceEvaluation):
    """Provides evaluation of a plane at given parameters.

    Parameters
    ----------
    plane: ~ansys.geometry.core.shapes.surfaces.plane.PlaneSurface
        Plane to evaluate.
    parameter: ParamUV
        Parameters (u, v) to evaluate the plane at.
    """

    def __init__(self, plane: PlaneSurface, parameter: ParamUV) -> None:
        """``SphereEvaluation`` class constructor."""
        self._plane = plane
        self._parameter = parameter

    @property
    def plane(self) -> PlaneSurface:
        """Plane being evaluated."""
        return self._plane

    @property
    def parameter(self) -> ParamUV:
        """Parameter that the evaluation is based upon."""
        return self._parameter

    @cached_property
    def position(self) -> Point3D:
        """Point on the surface, based on the evaluation."""
        return (
            self.plane.origin
            + self.parameter.u * self.plane.dir_x
            + self.parameter.v * self.plane.dir_y
        )

    @cached_property
    def normal(self) -> UnitVector3D:
        """Normal to the surface."""
        return self.plane.dir_z

    @cached_property
    def u_derivative(self) -> Vector3D:
        """First derivative with respect to u."""
        return self.plane.dir_z

    @cached_property
    def v_derivative(self) -> Vector3D:
        """First derivative with respect to v."""
        return self.plane.dir_y

    @cached_property
    def uu_derivative(self) -> Vector3D:
        """Second derivative with respect to u."""
        return Vector3D([0, 0, 0])

    @cached_property
    def uv_derivative(self) -> Vector3D:
        """Second derivative with respect to u and v."""
        return Vector3D([0, 0, 0])

    @cached_property
    def vv_derivative(self) -> Vector3D:
        """Second derivative with respect to v."""
        return Vector3D([0, 0, 0])

    @cached_property
    def min_curvature(self) -> Real:
        """Minimum curvature."""
        return 0

    @cached_property
    def min_curvature_direction(self) -> UnitVector3D:
        """Minimum curvature direction."""
        return self.plane.dir_x

    @cached_property
    def max_curvature(self) -> Real:
        """Maximum curvature."""
        return 0

    @cached_property
    def max_curvature_direction(self) -> UnitVector3D:
        """Maximum curvature direction."""
        return self.plane.dir_y



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a face (closed 2D sketch)."""

from typing import TYPE_CHECKING

from pint import Quantity
import pyvista as pv

from ansys.geometry.core.sketch.edge import SketchEdge

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.math.plane import Plane


class SketchFace:
    """Provides for modeling a face."""

    def __init__(self):
        """Initialize the face."""
        # TODO: What about the circular faces? Circle, Ellipse are not making use of this...
        # https://github.com/ansys/pyansys-geometry/issues/1319
        self._edges = []

    @property
    def edges(self) -> list[SketchEdge]:
        """List of all component edges forming the face."""
        return self._edges

    @property
    def perimeter(self) -> Quantity:
        """Perimeter of the face."""
        perimeter = 0
        for edge in self._edges:
            perimeter += edge.length
        return perimeter

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        meshes = [edge.visualization_polydata for edge in self.edges]
        return pv.merge(meshes)

    def plane_change(self, plane: "Plane") -> None:
        """Redefine the plane containing ``SketchFace`` objects.

        Notes
        -----
        This implies that their 3D definition might suffer changes. This method does
        nothing by default. It is required to be implemented in child ``SketchFace`` classes.

        Parameters
        ----------
        plane : Plane
            Desired new plane that is to contain the sketched face.
        """
        pass



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Auxiliary functions for the PyAnsys Geometry library."""

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.component import Component
    from ansys.geometry.core.designer.design import Design
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face


def get_design_from_component(component: "Component") -> "Design":
    """Get the ``Design`` of the given ``Component`` object.

    Parameters
    ----------
    component : Component
        The component object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided component object.
    """
    # Get the design of the component
    while component.parent_component is not None:
        component = component.parent_component

    # Return the design
    return component


def get_design_from_body(body: "Body") -> "Design":
    """Get the ``Design`` of the given ``Body`` object.

    Parameters
    ----------
    body : Body
        The body object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided body object.
    """
    # Get the parent component of the body
    component = body.parent_component

    # Get the design of the component
    return get_design_from_component(component)


def get_design_from_face(face: "Face") -> "Design":
    """Get the ``Design`` of the given ``Face`` object.

    Parameters
    ----------
    face : Face
        The face object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided face object.
    """
    # Get the parent body of the face
    body = face.body

    # Get the design of the body
    return get_design_from_body(body)


def get_design_from_edge(edge: "Edge") -> "Design":
    """Get the ``Design`` of the given ``Edge`` object.

    Parameters
    ----------
    edge : Edge
        The edge object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided edge object.
    """
    # Get the one of the bodies of the edge
    body = edge.faces[0].body

    # Get the design of the body
    return get_design_from_body(body)


def __traverse_all_bodies(comp: Union["Design", "Component"]) -> list["Body"]:
    """Traverse all bodies in a design and all its subcomponents.

    This is a private method. Do not use it directly.

    Notes
    -----
    This method is a recursive helper function to traverse all bodies in a
    design and all its subcomponents.

    Parameters
    ----------
    design : Design
        Design object to traverse.

    Returns
    -------
    list[Body]
        List of all bodies in the design or component.
    """
    bodies = []
    bodies.extend(comp.bodies)
    for component in comp.components:
        bodies.extend(__traverse_all_bodies(component))

    return bodies


def get_bodies_from_ids(design: "Design", body_ids: list[str]) -> list["Body"]:
    """Find the ``Body`` objects inside a ``Design`` from its ids.

    Notes
    -----
    This method takes a design and body ids, and gets their corresponding ``Body`` object.

    Parameters
    ----------
    design : Design
        Parent design for the faces.
    body_ids : list[str]
        List of body ids.

    Returns
    -------
    list[Body]
        List of Body objects.
    """
    return [body for body in __traverse_all_bodies(design) if body.id in body_ids]


def get_faces_from_ids(design: "Design", face_ids: list[str]) -> list["Face"]:
    """Find the ``Face`` objects inside a ``Design`` from its ids.

    Notes
    -----
    This method takes a design and face ids, and gets their corresponding ``Face`` object.

    Parameters
    ----------
    design : Design
        Parent design for the faces.
    face_ids : list[str]
        List of face ids.

    Returns
    -------
    list[Face]
        List of Face objects.
    """
    return [
        face for body in __traverse_all_bodies(design) for face in body.faces if face.id in face_ids
    ]  # noqa: E501


def get_edges_from_ids(design: "Design", edge_ids: list[str]) -> list["Edge"]:
    """Find the ``Edge`` objects inside a ``Design`` from its ids.

    Notes
    -----
    This method takes a design and edge ids, and gets their corresponding ``Edge`` objects.

    Parameters
    ----------
    design : Design
        Parent design for the edges.
    edge_ids : list[str]
        List of edge ids.

    Returns
    -------
    list[Edge]
        List of Edge objects.
    """
    return [
        edge for body in __traverse_all_bodies(design) for edge in body.edges if edge.id in edge_ids
    ]  # noqa: E501
